/*
**
** shaperd $VER: 1.33
**
** (C) 2002 Grzegorz Fitrzyk SP9WUN
**   email: sp9wun@poczta.onet.pl
**    http://sp9wun.republika.pl/
**
*/

#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <math.h>
#include <syslog.h>
#include <signal.h>
#include <setjmp.h>

#define TIMEOUT		300
#define IP_CONNTRACK 	"/proc/net/ip_conntrack"
#define PIDFILE		"/var/run/shaperd.pid"
#define CONFIGFILE	"/etc/shaper/shaper.cfg"
#define IGNOREFILE	"/etc/shaper/ignore"
#define IPLISTFILE	"/etc/shaper/iplist"
#define OLDFILE		"/var/shaper/bitrate_user_sh.old"
#define DEVFILE		"/proc/net/dev"
#define VERSION		"1.33 (06.09.2002)"

int			exit_signal = 0;
jmp_buf   		env_buf;

char			***natstat_tab_tmp;
int			natstat_tab_pos = 0;
char			**iplista;
char			**iplista_iface;
unsigned long int	*iplista_lo;
unsigned long int	*iplista_hi;
unsigned long int	*ipspeed;
unsigned long int       *iptime;
unsigned long int	*ipspeed_last;
unsigned long int       *iptime_last;
int			iplista_pos = 0;
int			**local_int_tab2 = NULL;
int			local_int_pos = 0;
int			bitrate_user_last_pos = 0;
int			ignore_pos = 1;
char			**devlist;
int			devlist_pos = 1;
int			masqip_pos = 1;

// Config
char			mainip[20] = "0.0.0.0";
char			internet_iface[12] = "ppp0";
unsigned long int       internet_iface_speed = 140000;
int			upload_control = 0;
int			high_start_speed = 1;
int			squid = 0;
unsigned long int	lospeed = 8000;
unsigned long int	hispeed = 100000;
char			speed_ext[10] = "\0";
int			delay	= 10;
int			olddelay= 6;
int			debug = 0;
int			notc = 0;
int			check_always = 0;
void frmem(void);
void error(char *, ...);
int my_system (const char *);
int numlines(char *);
unsigned long int ckbitrate_user(char *);
void read_iplist(void);
void read_firewall(int);
void write_bitrate(char *);
char **dev_list(void);
void cbq_off(char *);
void cbq_off_all(char **);
void cbq_setup_root(char *, unsigned long int, int);
void cbq_setup_clas(char *, unsigned long int, int, char *, char *);
void cbq_class_add(char *, unsigned long int, unsigned long int, int, int, char *, int , int, char *);
void cbq_standard_on(char ***);
char ***read_config(char *);
void free_config(char ***);
char ***natstat(char **, char ***, int);
void free_natstat_tab(char ***);
char **read_ignore(char *);
void free_ignore(char **);
char **get_masqip(char ***);
void free_masqip(char **);
char ***create_speed_table(int);
void free_speed_table(char ***, int);
void add_ip2speed_table_new(char ***,char *);
void add_ip2speed_table_sub1(char ***,char *);
void add_ip2speed_table_nop1(char ***,char *);
void add_ip2speed_table_add1(char ***,char *);
void add_ip2speed_table_sub2(char ***,char *);
void add_ip2speed_table_nop2(char ***,char *);
void add_ip2speed_table_add2(char ***,char *);
char *get_speed_table_new(char ***, int);
char *get_speed_table_sub1(char ***, int);
char *get_speed_table_nop1(char ***, int);
char *get_speed_table_add1(char ***, int);
char *get_speed_table_sub2(char ***, int);
char *get_speed_table_nop2(char ***, int);
char *get_speed_table_add2(char ***, int);
char *get_ip_iface(char *);
unsigned long int get_ip_lospeed(char *);
unsigned long int get_ip_hispeed(char *);
int get_config_by_ip(char ***, char *);
char *check_mask(char *);
int ip2hex(char *);
int ip2hex2(char *);
char *hex2ip(int);
char *hex2ip2(char *);
int mask2hex(char *);
char *get_ip_iface2(char ***, char *);
int mask2hex(char *);
void create_firewall(char ***, int);
void delete_firewall(char ***, int);
char ***mkbitrate_user_last(char ***, char *, long int);
char ***clear_bitrate_user_last(char ***, char **);
long int ckbitrate_user_last(char ***, char *);
void free_bitrate_user_last(char ***);
void write_bitrate_user_last(char ***,char *);
static int daemonize(void);
int check_linux(void);
int check_ipchains(void);
int check_iptables(void);
int is_shaperd_running(void);
int is_iproute2(void);
int is_qos(void);
void shownat(char ***);
char *lower(char *);
void exit_handler();



int main(int argc, char *argv[])
{
    int			n,nr,mult;
    char		***nattab;
    char		***config;
    char		***speed_tab;
    char		***bitrate_user_last = NULL;
    char		**lista;
    char		**ignore;
    char		**masq;
    char		*str;
    unsigned long int	speed;
    unsigned long int	tspeed;
    unsigned long int	reszta;
    unsigned long int   currbitrate;
    unsigned long int   currbitrate_total;
    long int		lastbitrate;
    unsigned long int   speed_out;	// predkosc upload
    int			ilip;
    int			cnt;
    int			log;
    int			old = olddelay;
    int			ipcol = 0;
    int			firewall_type = 0;
    int			start_time;
    int			delta_time;
    int			delay_org;

    log = 1;

    // Start syslog
    if (log)
    {
        openlog("shaperd", LOG_CONS|LOG_PERROR, LOG_DAEMON);
	openlog("shaperd", LOG_CONS, LOG_DAEMON);
    }

    if ((int) signal(SIGTERM, exit_handler) == -1)
    {
	syslog(LOG_ERR,"Cannot setup SIGTERM handler.");
	syslog(LOG_ERR,"Exit.");
	exit(1);
    }

    // Sprawdzamy czy pakiet iproute2 jest zainstalowany
    if (!is_iproute2())
    {
	syslog(LOG_ERR,"Package iproute2 not found.");
	syslog(LOG_ERR,"Exit.");
	exit(1);
    }
    
    // Sprawdzamy czy w kernelu jest obsluga QoS
    if (!is_qos())
    {
	syslog(LOG_ERR,"This kernel don't supporte QoS - recompile it wirh QoS support enabled.");
	syslog(LOG_ERR,"Exit.");
	exit(1);
    }

    // Sprawdzamy rodzaj firewalla
    if (check_ipchains())
    {
	if (debug) syslog(LOG_INFO,"Checking firewall type: ipchains");
	firewall_type=1;
    }
    else
    if (check_iptables())
    {
	if (debug) syslog(LOG_INFO,"Checking firewall type: iptables");
	firewall_type=2;
    }

    if (!firewall_type)
    {
	syslog(LOG_ERR,"Unknown firewall type");
	syslog(LOG_ERR,"Exit.");
	exit(1);
    }
    
    // czytamy plik konfiguracyjny
    config = read_config(CONFIGFILE);

    // sprawdzamy parametry podane przy starcie programu
    if (argc>1)
    {
	char	*p;
	int	fin;

	for(n=1;n<argc;n++)
	{
	    fin = 0;
	    p = strtok(argv[n],"=");
	    if (!strcmp(p,"-lospeed"))
	    {
		p = strtok(NULL," ");
		lospeed = strtoul(p,NULL,10);
		fin = 1;
	    }
	    else
	    if (!strcmp(p,"-hispeed"))
	    {
		p = strtok(NULL," ");
		hispeed = strtoul(p,NULL,10);
		fin = 1;
	    }
	    else
	    if (!strcmp(p,"-speed_ext"))
	    {
		p = strtok(NULL," ");
		if (! strcmp(p,"Kbit") || ! strcmp(p,"Mbit")) strncpy(speed_ext,p,10);
		else strncpy(speed_ext,"\0",10);
		fin = 1;
	    }
	    else
	    if (!strcmp(p,"-debug"))
	    {
		p = strtok(NULL," ");
		debug = atoi(p);
		fin = 1;
	    }
	    else
	    if (!strcmp(p,"-delay"))
	    {
		p = strtok(NULL," ");
		delay = atoi(p);
		if (delay < 10) delay = 10;
		fin = 1;
	    }
	    else
	    if (!strcmp(p,"-write_delay"))
	    {
		p = strtok(NULL," ");
		olddelay = atoi(p);
		if (olddelay < 0) olddelay = 0;
		fin = 1;
	    }
	    else
	    if (!strcmp(p,"-high_start_speed"))
	    {
		p = strtok(NULL," ");
		high_start_speed = atoi(p);
		fin = 1;
	    }
	    else
	    if (!strcmp(p,"-check_always"))
	    {
		p = strtok(NULL," ");
		check_always = atoi(p);
		fin = 1;
	    }
	    else
	    if (!strcmp(p,"-squid_support"))
	    {
		p = strtok(NULL," ");
		squid = atoi(p);
		fin = 1;
		if (squid) syslog(LOG_INFO,"squid_support active");
	    }
	    else
	    if (!strcmp(p,"-internet_iface"))
	    {
		p = strtok(NULL," ");
		strncpy(internet_iface,p,10);
		fin = 1;
	    }
	    else
	    if (!strcmp(p,"-internet_iface_speed"))
	    {
		p = strtok(NULL," ");
		internet_iface_speed = strtoul(p,NULL,10);
		if (internet_iface_speed <= hispeed)
		{
		    upload_control = 0;
		    syslog(LOG_ERR,"internet_iface_speed amount is less than hispeed. Deactivate Upload Bandwidth Limit");
		}
		else
		{
		    upload_control = 1;
		    syslog(LOG_INFO,"Upload Bandwidth Limit active");
		}
		fin = 1;
	    }
	    else
	    if (!strcmp(p,"-mainip"))
	    {
		p = strtok(NULL," ");
		strncpy(mainip,p,15);
		fin = 1;
	    }
	    else
	    if (!strcmp(p,"shownat"))
	    {
		ignore = read_ignore(IGNOREFILE);
		read_iplist();
		lista = dev_list();
		nattab = natstat(ignore, config, firewall_type);
		shownat(nattab);
		closelog();
		free_natstat_tab(nattab);
	        free_ignore(ignore);
		free_config(config);
		frmem();
		exit(0);
	    }
	    else
	    if (!strcmp(p,"--help"))
	    {
		printf("Usage: shaperd [options] [commands]
Options:
\t-mainip=IP number of internet interface
\t-high_start_speed=0|1 high speed start switch (deflt. 1)
\t-check_always=0|1 always control bandwidth (deflt. 0)
\t-squid_support=0|1 PROXY support switch (deflt. 0)
\t-lospeed guaraneed minimum download speed (global) (deflt. 8000)
\t-hispeed maximum download speed (global) (deflt. 100000)
\t-internet_iface=ppp0 internet interface name (deflt. ppp0)
\t-internet_iface_speed max. internet link speed (up+down) (deflt. 140000)
\t-speed_ext=bit|Kbit|Mbit jednostka szybkosci (deflt. bit)
\t-delay=seconds cycle delay time (deflt 10)
\t-write_delay=n after how many cycles to write bitrate_user_sh.old file\n\t\t\t(domysl. 6)
\t-debug=0|1|2 debug level (deflt. 0)
\t--help help

Commands:
\tshownat - show current masqueraded connections (without ignore)\n\n");

		closelog();
		free_config(config);
		frmem();
		exit(1);
	    }
	    else
	    if (fin == 0)
	    {
		printf("shaperd: syntax error. Unknown parameter: %s.\ntry: shaperd --help\n\n",argv[n]);
		syslog(LOG_ERR,"syntax error. Unknown parameter: %s.",argv[n]);
		closelog();
		free_ignore(ignore);
		free_config(config);
		frmem();
		exit(1);
	    }
	}
    }

    // sprawdzamy czy juz jakis shaper nie pracuje
    if (is_shaperd_running())
    {
	syslog(LOG_ERR,"Any shaperd is still working.");
	closelog();
	free_config(config);
	frmem();
	exit(1);
    }

    // czytamy liste ignorowanych polaczen
    ignore = read_ignore(IGNOREFILE);
    
    // czytamy liste lokalnych numerow ip
    read_iplist();

    // czytamy liste interfejsow w systemie
    lista = dev_list();

    // uruchamiamy program jako daemona systemowego
    if (daemonize() < 0) exit(1);

    // tworzymy regulki na firewallu rejestrujace ilosc pobranych danych
    delete_firewall(config,firewall_type);
    create_firewall(config,firewall_type);

    syslog(LOG_INFO, "Initialisation finished");

    delay_org = delay;

    mult = 1;
    if (! strcmp(speed_ext,"bit")) mult = 1;
    else if (! strcmp(speed_ext,"Kbit")) mult = 1024;
    else if (! strcmp(speed_ext,"Mbit")) mult = 1048576;

    // petla glowna
    while(!exit_signal)
    {
	start_time = time(NULL);
	currbitrate_total = 0;

	// czytamy na firewallu ilosc pobranych danych 
	read_firewall(firewall_type);

	// czytamy wszystkie polaczenia na maskaradzie
	nattab = natstat(ignore, config, firewall_type);

	// tworzymy liste maskowanych numerow ip dla przydzialow
	masq = get_masqip(nattab);
	nr = masqip_pos;
	if (debug) printf("Number of IP's for allocation bandwidth: %d\n",nr-1);

	// tworzymy tablice z aktualnymi przydzialami
        speed_tab = create_speed_table(nr);
	
	// czyscimy tablice z ostatnimi przydzialami
	bitrate_user_last = clear_bitrate_user_last(bitrate_user_last,masq);

	// robimy ograniczenia tylko wtedy gdy jest wiecej niz jeden user
        if (nr-1 > 1 || (nr-1 > 0 && check_always))
	{
	    if (debug) printf("ip: IP\t\t\tCurr.Bitrate\tLast Bitrate\n");
	    ilip = nr-1;
	    speed = hispeed / ilip;
	    for(n=1;n<nr;n++)
	    {
		currbitrate = ckbitrate_user(masq[n]);
		currbitrate_total = currbitrate_total + currbitrate;
		lastbitrate = ckbitrate_user_last(bitrate_user_last,masq[n]) * mult;
		if (lastbitrate < 0 || lastbitrate == 0)
		    add_ip2speed_table_new(speed_tab,masq[n]);
		else
		if (lastbitrate <= speed)
		{
	    	    if (currbitrate < lastbitrate*0.5)
			add_ip2speed_table_sub1(speed_tab,masq[n]);
		    else
		    if (currbitrate < lastbitrate*0.7)
			add_ip2speed_table_nop1(speed_tab,masq[n]);
		    else
			add_ip2speed_table_add1(speed_tab,masq[n]);
		}
		else
		{
		    if (currbitrate < lastbitrate*0.5)
			add_ip2speed_table_sub2(speed_tab,masq[n]);
		    else
		    if (currbitrate < lastbitrate*0.7)
			add_ip2speed_table_nop2(speed_tab,masq[n]);
		    else
			add_ip2speed_table_add2(speed_tab,masq[n]);
		}
		if (debug) printf("ip: %s\t\t%lu\t\t%ld\n",masq[n],currbitrate,lastbitrate);
	    }

	    // usuwamy wszystkie ograniczenia
	    cbq_off_all(lista);

	    // ustawiamy standardowe ograniczenia
	    cbq_standard_on(config);
	    reszta = hispeed;
	    cnt = 4;

	    for(n=1;n<nr;n++)
	    {
		int		pos;
		char		*intf;
		int		bcn,isp;
		int		his;

    		str = get_speed_table_new(speed_tab,n);
		if (str[0])
		{
		    if (high_start_speed)
		    {
			his = get_ip_hispeed(str);
			tspeed = speed;
			if (tspeed > his) tspeed = his;
			if (debug) printf("high_speed_start: %ld\n",tspeed);
		    }
		    else tspeed = get_ip_lospeed(str);

    		    if ((pos = get_config_by_ip(config,str)) != -1)
		    {
		        intf = get_ip_iface(str);
			bcn = atoi(config[1][(pos*6)+5]);
			isp = atoi(config[1][(pos*6)+2]);
			cbq_class_add(intf,isp,tspeed,bcn,cnt,str,3,20,"bounded");
		        bitrate_user_last = mkbitrate_user_last(bitrate_user_last,str,tspeed);
		        reszta = reszta - tspeed;
		        ilip--;
		        cnt++;
			if (cnt > 255) syslog(LOG_ERR,"Ooops! Too many IP numbers. Could by problem with creating CBQ class.");
		    }
		    else
		    {
	    	        syslog(LOG_ERR,"Interface name for IP number %s not founf in /etc/shaper/iplist",str);
		    }
		    if (debug) printf("new:  %s\n",str);
		}
	    }

	    for(n=1;n<nr;n++)
	    {
	        int		pos;
	        char		*intf;
	        int		bcn,isp;
		int		los;

		str = get_speed_table_sub1(speed_tab,n);
		if (str[0])
		{
	            lastbitrate = ckbitrate_user_last(bitrate_user_last,str) * mult;
	            tspeed = (lastbitrate * 0.8) / mult;
		    los = get_ip_lospeed(str);
	            if (tspeed < los) tspeed = los;
    	            if ((pos = get_config_by_ip(config,str)) != -1)
		    {
	    	        intf = get_ip_iface(str);
		        bcn = atoi(config[1][(pos*6)+5]);
		        isp = atoi(config[1][(pos*6)+2]);
		        cbq_class_add(intf,isp,tspeed,bcn,cnt,str,2,20,"isolated");
		        bitrate_user_last = mkbitrate_user_last(bitrate_user_last,str,tspeed);
		        reszta = reszta - tspeed;
		        ilip--;
		        cnt++;
			if (cnt > 255) syslog(LOG_ERR,"Ooops! Too many IP numbers. Could by problem with creating CBQ class.");
		    }
		    else
	    	    {
	    	        syslog(LOG_ERR,"Interface name for IP number %s not founf in /etc/shaper/iplist",str);
		    }
		    if (debug) printf("sub1:  %s\n",str);
		}
	    }

	    for(n=1;n<nr;n++)
	    {
	        int		pos;
	        char		*intf;
	        int		bcn,isp;
		int		his;

		str = get_speed_table_nop1(speed_tab,n);
	        if (str[0])
	        {
	    	    lastbitrate = ckbitrate_user_last(bitrate_user_last,str) * mult;
		    his = get_ip_hispeed(str);
		    tspeed = lastbitrate / mult;
		    if (tspeed > his) tspeed = his;
    		    if ((pos = get_config_by_ip(config,str)) != -1)
		    {
		        intf = get_ip_iface(str);
		        bcn = atoi(config[1][(pos*6)+5]);
		        isp = atoi(config[1][(pos*6)+2]);
		        cbq_class_add(intf,isp,tspeed,bcn,cnt,str,3,20,"isolated");
		        bitrate_user_last = mkbitrate_user_last(bitrate_user_last,str,tspeed);
		        reszta = reszta - tspeed;
		        ilip--;
		        cnt++;
			if (cnt > 255) syslog(LOG_ERR,"Ooops! Too many IP numbers. Could by problem with creating CBQ class.");
		    }
		    else
		    {
		        syslog(LOG_ERR,"Interface name for IP number %s not founf in /etc/shaper/iplist",str);
		    }
		    if (debug) printf("nop1:  %s\n",str);
		}
	    }


	    for(n=1;n<nr;n++)
	    {
	        int		pos;
	        char		*intf;
		int		bcn,isp;
		int		los,his;

	        str = get_speed_table_add1(speed_tab,n);
		if (str[0])
	        {
	    	    lastbitrate = ckbitrate_user_last(bitrate_user_last,str) * mult;
		    los = get_ip_lospeed(str);
		    his = get_ip_hispeed(str);
		    tspeed = ((lastbitrate/mult) + los);
		    if (tspeed > (reszta/ilip))
			tspeed = reszta / ilip;
		    if (tspeed > his) tspeed = his;
    		    if ((pos = get_config_by_ip(config,str)) != -1)
		    {
	    	        intf = get_ip_iface(str);
		        bcn = atoi(config[1][(pos*6)+5]);
		        isp = atoi(config[1][(pos*6)+2]);
		        cbq_class_add(intf,isp,tspeed,bcn,cnt,str,4,20,"isolated");
			bitrate_user_last = mkbitrate_user_last(bitrate_user_last,str,tspeed);
		        reszta = reszta - tspeed;
		        ilip--;
		        cnt++;
			if (cnt > 255) syslog(LOG_ERR,"Ooops! Too many IP numbers. Could by problem with creating CBQ class.");
		    }
		    else
		    {
		        syslog(LOG_ERR,"Interface name for IP number %s not founf in /etc/shaper/iplist",str);
		    }
		    if (debug) printf("add1:  %s\n",str);
		}
    	    }

	    for(n=1;n<nr;n++)
	    {
    	        int		pos;
		char		*intf;
	        int		bcn,isp;
		int		los;

	        str = get_speed_table_sub2(speed_tab,n);
	        if (str[0])
	        {
		    lastbitrate = ckbitrate_user_last(bitrate_user_last,str) * mult;
		    tspeed = (lastbitrate * 0.8) / mult;
		    los = get_ip_lospeed(str);
	            if (tspeed < los) tspeed = los;
    	    	    if ((pos = get_config_by_ip(config,str)) != -1)
	    	    {
	    		intf = get_ip_iface(str);
			bcn = atoi(config[1][(pos*6)+5]);
		        isp = atoi(config[1][(pos*6)+2]);
		        cbq_class_add(intf,isp,tspeed,bcn,cnt,str,5,5,"bounded");
			bitrate_user_last = mkbitrate_user_last(bitrate_user_last,str,tspeed);
		        reszta = reszta - tspeed;
		        ilip--;
		        cnt++;
			if (cnt > 255) syslog(LOG_ERR,"Ooops! Too many IP numbers. Could by problem with creating CBQ class.");
		    }
	    	    else
		    {
	    	        syslog(LOG_ERR,"Interface name for IP number %s not founf in /etc/shaper/iplist",str);
	    	    }
	    	    if (debug) printf("sub2:  %s\n",str);
		}
	    }

	    for(n=1;n<nr;n++)
    	    {
	        int		pos;
	        char		*intf;
		int		bcn,isp;
		int		los,his;

		str = get_speed_table_nop2(speed_tab,n);
	        if (str[0])
	        {
	            lastbitrate = ckbitrate_user_last(bitrate_user_last,str) * mult;
	            tspeed = lastbitrate / mult;
		    his = get_ip_hispeed(str);
	            if (tspeed > (reszta/ilip))
	    	        tspeed = reszta/ilip;
		    if (tspeed > his) tspeed = his;
    	    	    if ((pos = get_config_by_ip(config,str)) != -1)
	    	    {
			intf = get_ip_iface(str);
		        bcn = atoi(config[1][(pos*6)+5]);
		        isp = atoi(config[1][(pos*6)+2]);
		        cbq_class_add(intf,isp,tspeed,bcn,cnt,str,6,2,"bounded");
		        bitrate_user_last = mkbitrate_user_last(bitrate_user_last,str,tspeed);
		        reszta = reszta - tspeed;
		        ilip--;
		        cnt++;
			if (cnt > 255) syslog(LOG_ERR,"Ooops! Too many IP numbers. Could by problem with creating CBQ class.");
		    }
	    	    else
	    	    {
	    	        syslog(LOG_ERR,"Interface name for IP number %s not founf in /etc/shaper/iplist",str);
		    }
	    	    if (debug) printf("nop2:  %s\n",str);
		}
	    }


	    for(n=1;n<nr;n++)
    	    {
	        int		pos;
		char		*intf;
		int		bcn,isp;
		int		los,his;

		str = get_speed_table_add2(speed_tab,n);
	        if (str[0])
		{
		    lastbitrate = ckbitrate_user_last(bitrate_user_last,str) * mult;
		    los = get_ip_lospeed(str);
		    his = get_ip_hispeed(str);
		    tspeed = ((lastbitrate/mult) + los);
		    if (tspeed > (reszta/ilip))
			tspeed = reszta / ilip;
		    if (tspeed > his) tspeed = his;
    		    if ((pos = get_config_by_ip(config,str)) != -1)
	    	    {
			intf = get_ip_iface(str);
		        bcn = atoi(config[1][(pos*6)+5]);
		        isp = atoi(config[1][(pos*6)+2]);
			cbq_class_add(intf,isp,tspeed,bcn,cnt,str,6,1,"bounded");
		        bitrate_user_last = mkbitrate_user_last(bitrate_user_last,str,tspeed);
		        reszta = reszta - tspeed;
		        ilip--;
		        cnt++;
			if (cnt > 255) syslog(LOG_ERR,"Ooops! Too many IP numbers. Could by problem with creating CBQ class.");
		    }
		    else
		    {
		        syslog(LOG_ERR,"Interface name for IP number %s not founf in /etc/shaper/iplist",str);
		    }
		    if (debug) printf("add2:  %s\n",str);
		}
	    }

	    // ograniczamy szybkosc upload
	    if (upload_control)
	    {
//		speed_out = internet_iface_speed + (reszta - hispeed);
		if (! strcmp(speed_ext,"bit")) speed_out = internet_iface_speed - currbitrate_total;
		else if (! strcmp(speed_ext,"Kbit")) speed_out = internet_iface_speed - (currbitrate_total / 1024);
		else if (! strcmp(speed_ext,"Mbit")) speed_out = internet_iface_speed - (currbitrate_total / 1048576);
		if (speed_out > hispeed) speed_out = hispeed;
		if (debug) printf("upload speed: %lu\n",speed_out);
		cbq_setup_root(internet_iface,speed_out,255);
		cbq_setup_clas(internet_iface,speed_out,255,"0.0.0.0/0","0.0.0.0/0");
	    }
	}
	else
	// dla jednego usera nie robimy ograniczen
	if (nr-1 < 1 || (!check_always && nr-1 <=1))
	{
	    // usuwamy wszystkie ograniczenia
	    cbq_off_all(lista);
	}

	// sprawdzamy czas przydzialu widelek, jesli za dlugi w stosunku do
	// opoznienia delay to zwiekszamy opoznienie o 10%
	// jesli czas jest bardzo krotki to zmniejszamy opoznienie.
	delta_time = time(NULL)-start_time;
	if (debug) printf("Time of creating allocations %d\n",delta_time);

	if (delta_time>delay * 0.4)
	{
	    delay = delay * 1.25;
    	    syslog(LOG_ERR,"Delay parameter has too low amount for this number of IP's or possibility of server.");
	    syslog(LOG_ERR,"Increasing delay to %d seconds",delay);
	}
	else
	if (delta_time<delay * 0.25)
	{
	    delay = delay * 0.75;
	    if (delay <= delay_org) delay = delay_org;
	    else
	    {
		syslog(LOG_ERR,"delay parameter has too high amount.");
		syslog(LOG_ERR,"Decreasing delay to %d seconds",delay);
	    }
	}

	n=0;
	start_time = time(NULL);
	while(!exit_signal && n<delay)
	{
	    sleep(1);
	    n++;
	}
	delta_time = time(NULL)-start_time;
	if (debug) printf("Delay time %d\n",delta_time);

	// zapisujemy plik bitrate_user_sh.old
	old++;
	if (old >= olddelay)
	{
	    if (nr-1 == 1 && !check_always)
	    {
		bitrate_user_last = clear_bitrate_user_last(bitrate_user_last,masq);
		bitrate_user_last = mkbitrate_user_last(bitrate_user_last,masq[1],hispeed);
	    }
	    else
	    if (nr-1 < 1)
	    {
		bitrate_user_last = clear_bitrate_user_last(bitrate_user_last,masq);
//		bitrate_user_last = mkbitrate_user_last(bitrate_user_last,"0.0.0.0",hispeed);
	    }
    	    write_bitrate_user_last(bitrate_user_last,OLDFILE);
	    old = 0;
	}
	free_masqip(masq);       
        free_natstat_tab(nattab);
        free_speed_table(speed_tab,nr);
    }
    // usuwamy wszystkie ograniczenia
    cbq_off_all(lista);              

    // usuwamy PID
    remove(PIDFILE);

    closelog();
    free_bitrate_user_last(bitrate_user_last);
    delete_firewall(config,firewall_type);
    free_ignore(ignore);
    free_config(config);
    frmem();
    exit(0);
}

void frmem(void)
{
    int		m;

    if (iplista)
    {
	m = 0;
	while(iplista[m])
	{
	    free(iplista[m]);
	    free(iplista_iface[m++]);
	}
	free(iplista);
    }
    if (iplista_lo) free(iplista_lo);
    if (iplista_hi) free(iplista_hi);
    if (ipspeed) free(ipspeed);
    if (ipspeed_last) free(ipspeed_last);
    if (iptime) free(iptime);
    if (iptime_last) free(iptime_last);
    if (devlist)
    {
	for(m=1;m<devlist_pos;m++)
	{
	    free(devlist[m]);
	}
	free(devlist);
    }
    return;
}

void error(char *fmt, ...)
{
    va_list	args;

    va_start(args, fmt);
    fprintf(stderr, "error: ");
    vfprintf(stderr, fmt, args);
    fprintf(stderr,"\n");
    va_end(args);

    frmem();

    exit(1);
}

int my_system (const char *command)
{
    char	*environ[]=
    {
	"HOME=/",
	"PATH=/sbin:/bin:/usr/sbin:/usr/bin",
	"TERM=linux",
	NULL
    };

    int pid, status;

    if (command == 0)
	return 1;
    pid = fork();
    if (pid == -1)
	return -1;
    if (pid == 0)
    {
	char *argv[4];

	argv[0] = "sh";
	argv[1] = "-c";
	argv[2] = (char *) command;
	argv[3] = 0;
	execve("/bin/sh", argv, environ);
	exit(127);
    }
    do
    {
	if (waitpid(pid, &status, 0) == -1)
	{
	    if (errno != EINTR) return -1;
	}
	else return status;
    }
    while(1);
}

int numlines(char *fname)
{
    FILE	*fp;
    int		line=0;
    char	buf[1000];
    
    if (!(fp = fopen(fname,"r")))
    {
	syslog(LOG_ERR,"File %s not found.",fname);
	exit(1);
    }
    while(fgets(buf,1000,fp))
	line++;
    fclose(fp);
    return(line);
}

unsigned long int ckbitrate_user(char *ip)
{
    int		m,n,pos;
    FILE	*fp;
    char	buf[1000];
    char	ip_tmp[20];
    char	ipt[20];
    long unsigned int	d1,d2,sp,tm,res;

    res = 0;    
    for(n=1;n<iplista_pos;n++)
    {
	if ((!strcmp(iplista[n],ip) && (!strcmp(ip,iplista[n]))))
	{
	    sp = ipspeed[n]-ipspeed_last[n];
	    tm = iptime[n]-iptime_last[n];
	    if (tm > 0)
	    {
		res = ((sp/tm)*8);
		ipspeed_last[n] = ipspeed[n];
		iptime_last[n] = iptime[n];
	    }
	}
    }
    return(res);
}

void read_iplist(void)
{
    int		f1,m,n;
    struct	stat stbuf;
    int		d1,d2,d3,d4;
    long unsigned int	d5,d6;
    char	*src;
    char	iface[10];
    char	buf[30];
    char	bufor[1000];
    FILE	*fp;

    m = numlines(IPLISTFILE) + 1;
    iplista = (char **) calloc(m,sizeof(char *));
    iplista_iface = (char **) calloc(m,sizeof(char *));
    iplista_lo = (long unsigned int *) calloc(m,sizeof(long unsigned int));
    iplista_hi = (long unsigned int *) calloc(m,sizeof(long unsigned int));
    ipspeed = (long unsigned int *) calloc(m,sizeof(long unsigned int));
    iptime  = (long unsigned int *) calloc(m,sizeof(long unsigned int));
    ipspeed_last = (long unsigned int *) calloc(m,sizeof(long unsigned int));
    iptime_last  = (long unsigned int *) calloc(m,sizeof(long unsigned int));

    iplista_pos = 1;
    fp = fopen(IPLISTFILE,"r");
    while ((fgets(bufor,1000,fp)) != NULL)
    {
	d5 = 0;
	d6 = 0;
	sscanf(bufor,"%3d.%3d.%3d.%3d=%s %lu %lu\n",&d1,&d2,&d3,&d4,iface,&d5,&d6);
	sprintf(buf,"%d.%d.%d.%d",d1,d2,d3,d4);
	iplista[iplista_pos] = (char *) calloc(strlen(buf)+2,sizeof(char));
	iplista_iface[iplista_pos] = (char *) calloc(strlen(iface)+2,sizeof(char));
	if (d5 == 0) d5 = lospeed;
	if (d5 > hispeed) d5 = hispeed;
	if (d6 == 0) d6 = hispeed;
	if (d6 > hispeed) d6 = hispeed;
	iplista_lo[iplista_pos] = d5;
	iplista_hi[iplista_pos] = d6;
	strcpy(iplista[iplista_pos],buf);
	strcpy(iplista_iface[iplista_pos],iface);
	if (debug) printf("read_iplist: %s %s %lu %lu\n",
	    iplista[iplista_pos],iplista_iface[iplista_pos],d5,d6);
	iplista_pos++;
    }
    fclose(fp);
}

void read_firewall(int firewall_type)
{
    int		f1,m,n;
    struct	stat stbuf;
    unsigned long int	cnt1,cnt2;
    unsigned long int	dd1,dd2,dd3,dd4;
    char	aa[12][20];
    char	*src;
    char	buf[30];
    char	buft[1000];

    FILE	*fw,*fp;

    fflush(NULL);
    if (firewall_type==1)
    {
	if (!(fp = popen("ipchains -vxL shaper -n 2> /dev/null","r")))
	{
	    syslog(LOG_ERR,"Can't read shaper table from ipchains");
//		error("Nie moge odczytac tablicy shaper w iptables");
	    exit(1);
	}
    }
    else
    if (firewall_type==2)
    {
	if (!(fp = popen("iptables -vxL shaper -n 2> /dev/null","r")))
	{
	    syslog(LOG_ERR,"Can't read shaper table from iptables");
//		error("Nie moge odczytac tablicy shaper w iptables");
	    exit(1);
	}
    }
    else
    {
	if (debug) syslog(LOG_ERR,"BUG: firewall type %d not exists!",firewall_type);
	exit(1);
    }

    while ((fgets(buft,1000,fp))!=NULL)
    {
	char	bufor[200];
	char	ipt[20];

	sscanf(buft,"%lu %lu %s %s %s %s %s %s %s %s %s",
	    &dd1,&dd2,aa[3],aa[4],aa[5],aa[6],aa[7],aa[8],aa[9],aa[10],aa[11]);
	for(m=1;m<iplista_pos;m++)
	{
	    strcpy(ipt,iplista[m]);
	    strcat(ipt," ");
	    if (strstr(buft,ipt))
	    {
		ipspeed[m] = dd2;
		iptime[m]  = (long unsigned int) time(NULL);
		if (debug > 1) printf("read_firewall: %s %lu %lu\n",iplista[m],ipspeed[m],iptime[m]);
	    }
	}
    }
    pclose(fp);
}

void write_bitrate(char *fname)
{
    FILE	*fw;
    int		m;
    int		fd;
    char	buf[200];

    fd = creat(fname,0644);
    for(m=1;m<iplista_pos;m++)
    {
	sprintf(buf,"%s %lu %lu\n",iplista[m],ipspeed[m],iptime[m]);
	write(fd,buf,strlen(buf));
    }
    close(fd);
}

char **dev_list(void)
{
    FILE        *fp;
    int         m,n;
    char	buft[1000];
    char	*buftptr=buft;
    char	iface[10];
    char	*devlist_cont;
    char	*res;

    m=numlines(DEVFILE);
    devlist = (char **) calloc(m+2,sizeof(char *));
    fp = fopen(DEVFILE,"r");
    fgets(buft,1000,fp);
    fgets(buft,1000,fp);
    while ((fgets(buft,1000,fp))!=NULL)
    {
	res=strtok(buft,":");
	while(isspace(*res))
	    res++;

	devlist[devlist_pos] = (char *) calloc(strlen(res)+2,sizeof(char));
	strcpy(devlist[devlist_pos],res);
	if (debug > 1) printf("devlist: %s\n",devlist[devlist_pos]);
	devlist_pos++;
    }
    fclose(fp);
    return(devlist);
}

void cbq_off(char *dev)
{
    char	buf[200];
    FILE	*fp;

    fflush(NULL);
    if (!notc && (fp = popen("/bin/sh 2>/dev/null","w")))
    {
        sprintf(buf,"tc qdisc del dev %s root 2>/dev/null\n",dev);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
	pclose(fp);
    }
}

void cbq_off_all(char **lista)
{
    int		n=1;

    if (debug > 0) printf("cbq_off_all\n");
    while(lista[n])
    {
	cbq_off(lista[n]);
	n++;
    }
}

void cbq_setup_root(char *iface, unsigned long int max_speed, int id1)
{
    unsigned long int	ts;
    char	buf[2000];
    FILE	*fp;

    fflush(NULL);
    if (!notc && (fp = popen("/bin/sh 2>/dev/null","w")))
    {
	ts = max_speed/id1;
	sprintf(buf,"tc qdisc add dev %s root handle %d: cbq bandwidth %lu%s avpkt 1000 cell 8\n",
		iface,id1,max_speed,speed_ext);
	if (debug > 0) printf("cbq_setup_root:\n");
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
        sprintf(buf,"tc class add dev %s parent %d:0 classid %d:1 cbq bandwidth %lu%s rate %lu%s allot 1514 cell 8 weight %lu%s prio 7 maxburst 20 avpkt 1000\n",
		iface,id1,id1,max_speed,speed_ext,max_speed,speed_ext,ts,speed_ext);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
	pclose(fp);
    }
}

void cbq_setup_clas(char *iface, unsigned long int max_speed, int id1, char *ipg, char *ipd)
{
    char        buf[2000];
    FILE	*fp;

    fflush(NULL);
    if (!notc && (fp = popen("/bin/sh 2>/dev/null","w")))
    {
	sprintf(buf,"tc class add dev %s parent %d:1 classid %d:2 cbq bandwidth %lu%s rate %lu%s allot 1514 cell 8 weight 1Kbit prio 5 maxburst 20 avpkt 1000 bounded\n",
	        iface, id1, id1, max_speed, speed_ext, max_speed, speed_ext);
	if (debug > 0) printf("cbq_setup_clas:\n");
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
        sprintf(buf,"tc qdisc add dev %s parent %d:2 tbf rate %lu%s buffer 10Kb/8 limit 15Kb\n",
	        iface, id1, max_speed, speed_ext);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);

	if (squid)
	{
	// jesli squid trafi to ustawi TOS=0 na pakietach z portu 3128
        sprintf(buf,"tc filter add dev %s parent %d:0 protocol ip prio 100 u32 match ip src %s match ip sport 3128 0xffff match ip tos 0 0x00 match ip dst %s flowid %d:2\n",
	        iface, id1, ipg, ipd, id1);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
	// porty od 0 do 3127 robimy bez reakcji na TOS
        sprintf(buf,"tc filter add dev %s parent %d:0 protocol ip prio 100 u32 match ip src %s match ip sport 0 0xf3c8 match ip dst %s flowid %d:2\n",
	        iface, id1, ipg, ipd, id1);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
	// porty od 3129 do 65535 robimy bez reakcji na TOS
        sprintf(buf,"tc filter add dev %s parent %d:0 protocol ip prio 100 u32 match ip src %s match ip sport 3129 0xf3c7 match ip dst %s flowid %d:2\n",
	        iface, id1, ipg, ipd, id1);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
	}
	else
	{
        sprintf(buf,"tc filter add dev %s parent %d:0 protocol ip prio 100 u32 match ip src %s match ip sport 0 0x0 match ip dst %s flowid %d:2\n",
	        iface, id1, ipg, ipd, id1);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
	}
/*
	sprintf(buf,"tc class add dev %s parent %d:1 classid %d:3 cbq bandwidth %lu%s rate %lu%s allot 1514 cell 8 weight 1Kbit prio 5 maxburst 20 avpkt 1000 bounded\n",
	        iface, id1, id1, max_speed, speed_ext, max_speed, speed_ext);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
        sprintf(buf,"tc qdisc add dev %s parent %d:3 tbf rate %lu%s buffer 10Kb/8 limit 15Kb\n",
	        iface, id1, max_speed, speed_ext);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
*/
	if (squid)
	{
        sprintf(buf,"tc filter add dev %s parent %d:0 protocol ip prio 100 u32 match ip src %s match ip sport 3128 0xffff match ip tos 0 0x00 match ip dst %s flowid %d:2\n",
		iface, id1, mainip, ipd, id1);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
        sprintf(buf,"tc filter add dev %s parent %d:0 protocol ip prio 100 u32 match ip src %s match ip sport 0 0xf3c8 match ip dst %s flowid %d:2\n",
		iface, id1, mainip, ipd, id1);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
        sprintf(buf,"tc filter add dev %s parent %d:0 protocol ip prio 100 u32 match ip src %s match ip sport 3129 0xf3c7 match ip dst %s flowid %d:2\n",
		iface, id1, mainip, ipd, id1);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
	}
	else
	{
	sprintf(buf,"tc filter add dev %s parent %d:0 protocol ip prio 100 u32 match ip src %s match ip sport 0 0x0 match ip dst %s flowid %d:2\n",
		iface, id1, mainip, ipd, id1);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
	}
	pclose(fp);
    }
}

void cbq_class_add(	char *iface, unsigned long int max_speed,
			unsigned long int t_speed, int id1, int id2,
			char *ipd, int pri, int maxburst, char *borrow)
{
    int		prio = pri + 100;
    char	buf[2000];
    FILE	*fp;

    fflush(NULL);
    if (!notc && (fp = popen("/bin/sh 2>/dev/null","w")))
    {
	sprintf(buf,"tc class add dev %s parent %d:1 classid %d:%d cbq bandwidth %lu%s rate %lu%s allot 1514 cell 8 weight 1Kbit prio %d maxburst %d avpkt 1000 %s\n",
    		iface,id1,id1,id2,max_speed,speed_ext,t_speed,speed_ext,pri,maxburst,borrow);
	if (debug > 0) printf("cbq_class_add:\n");
	if (debug > 1) printf("%s",buf); 
	fputs(buf,fp);
	sprintf(buf,"tc qdisc add dev %s parent %d:%d tbf rate %lu%s buffer 10Kb/8 limit 15Kb\n",
	        iface,id1,id2,t_speed,speed_ext);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
        sprintf(buf,"tc filter add dev %s parent %d:0 protocol ip prio %d u32 match ip sport 0 0x0 match ip dst %s flowid %d:%d\n",
    	        iface,id1,prio,ipd,id1,id2);
	if (debug > 1) printf("%s",buf);
	fputs(buf,fp);
	pclose(fp);
    }
}

void cbq_standard_on(char ***config)
{
    char    	*dev;
    int     	bcn;
    unsigned long int isp;
    char    	*ipg;
    char    	*ipu;
    int		n;

    for(n=0;n<local_int_pos;n++)
    {
	dev = config[1][(n*6)+1];
	bcn = atoi(config[1][(n*6)+5]);
	isp = strtoul(config[1][(n*6)+2],NULL,10);
	ipg = config[1][(n*6)+3];
	ipu = config[1][(n*6)+4];
	if (debug > 0)
	{
	    printf("cbq_standard_on:\n");
	    printf("cbq_setup_root:  %s %lu %d\n",dev,isp,bcn);
	    printf("cbq_setup_clas:  %s %lu %d %s %s\n",dev,isp,bcn,ipg,ipu);
	}
	cbq_setup_root(dev,isp,bcn);
	cbq_setup_clas(dev,isp,bcn,ipg,ipu);
    }
}

char ***read_config(char *fname)
{
    int		m,n;
    FILE	*fp;
    char	buf[1000];
    char	ip_tmp[20];
    char	ipt[20];
    long unsigned int	d1,d2,sp,tm;
    char	*res;
    char	***local_int_tab = NULL;
    char	**local_int;

    if (!(fp = fopen(fname,"r")))
    {
	syslog(LOG_ERR,"Can't read configuration file %s",fname);
//	error("Nie moge odczytac pliku konfiguracyjnego %s",fname);
	exit(1);
    }
//    printf("read_config: malloc local_int_tab\n");
    local_int_tab = (char ***) calloc(3,sizeof(char **));
    local_int_tab2 = (int **) calloc(3,sizeof(int *));
    local_int_tab[1] = (char **) NULL;
    local_int_tab2[1] = (int *) NULL;
    fgets(buf,1000,fp);
    res=strtok(buf,"=");
    if ((! strcmp("mainip",res)) && (! strcmp(res,"mainip")))
    {
	res = strtok(NULL,"\n");
	strncpy(mainip,res,15);
    }
    while ((fgets(buf,1000,fp))!=NULL)
    {
	res=strtok(buf,"=");
	if ((! strcmp("internet_iface",res)) && (! strcmp(res,"internet_iface")))
	{
//	    int		fin = 0;
	
	    res = strtok(NULL,"\n");
//	    for(n=1;n<devlist_pos;n++)
//		if (! strcmp(devlist[n],res) && ! strcmp(res,devlist[n]))
//		    fin = 1;
//	    if (fin) strncpy(internet_iface,res,10);
	    strncpy(internet_iface,res,10);
	}
	else
	if ((! strcmp("internet_iface_speed",res)) && (! strcmp(res,"internet_iface_speed")))
	{
	    int		fin = 0;
	
	    res = strtok(NULL,"\n");
	    internet_iface_speed = strtoul(res,NULL,10);
	    if (internet_iface_speed <= hispeed)
	    {
	        upload_control = 0;
	        syslog(LOG_ERR,"internet_iface_speed amount is less than hispeed. Limiting Upload Bandwidth deactivated.");
	    }
	    else
	    {
    		upload_control = 1;
		syslog(LOG_INFO,"Limiting Upload Bandwidth active");
	    }
	}
	else
	if ((! strcmp("high_start_speed",res)) && (! strcmp(res,"high_start_speed")))
	{
	    res = strtok(NULL,"\n");
	    high_start_speed = atoi(res);
	}
	else
	if ((! strcmp("check_always",res)) && (! strcmp(res,"check_always")))
	{
	    res = strtok(NULL,"\n");
	    check_always = atoi(res);
	}
	else
	if ((! strcmp("squid_support",res)) && (! strcmp(res,"squid_support")))
	{
	    res = strtok(NULL,"\n");
	    squid = atoi(res);
	    if (squid) syslog(LOG_INFO,"squid_support active");
	}
	else
	if ((! strcmp("lospeed",res)) && (! strcmp(res,"lospeed")))
	{
	    res = strtok(NULL,"\n");
	    lospeed = strtoul(res,NULL,10);
	}
	else
	if ((! strcmp("hispeed",res)) && (! strcmp(res,"hispeed")))
	{
	    res = strtok(NULL,"\n");
	    hispeed = strtoul(res,NULL,10);
	}
	else
	if ((! strcmp("speed_ext",res)) && (! strcmp(res,"speed_ext")))
	{
	    res = strtok(NULL,"\n");
	    if (! strcmp(res,"Kbit") || ! strcmp(res,"Mbit")) strncpy(speed_ext,res,10);
	    else strncpy(speed_ext,"\0",10);
	}
	else
	if ((! strcmp("debug",res)) && (! strcmp(res,"debug")))
	{
	    res = strtok(NULL,"\n");
	    debug = atoi(res);
	}
	else
	if ((! strcmp("notc",res)) && (! strcmp(res,"notc")))
	{
	    res = strtok(NULL,"\n");
	    notc = atoi(res);
	}
	else
	if ((! strcmp("delay",res)) && (! strcmp(res,"delay")))
	{
	    res = strtok(NULL,"\n");
	    delay = atoi(res);
	    if (delay < 10) delay = 10;
	}
	else
	if ((! strcmp("write_delay",res)) && (! strcmp(res,"write_delay")))
	{
	    res = strtok(NULL,"\n");
	    olddelay = atoi(res);
	    if (olddelay < 0) olddelay = 0;
	}
	else
	if ((! strcmp("local_int",res)) && (! strcmp(res,"local_int")))
	{
	    local_int_tab[1] = (char **) realloc((char **) local_int_tab[1],sizeof(char *)*6*(local_int_pos+1));
	    local_int_tab2[1] = (int *) realloc((int *) local_int_tab2[1],sizeof(int)*3*(local_int_pos+1));

	    res=strtok(NULL,";");
	    local_int_tab[1][(local_int_pos*6)+1]= (char *) calloc(strlen(res)+2,sizeof(char));
	    strcpy(local_int_tab[1][(local_int_pos*6)+1],res);

	    res=strtok(NULL,";");
	    local_int_tab[1][(local_int_pos*6)+2]= (char *) calloc(strlen(res)+2,sizeof(char));
	    strcpy(local_int_tab[1][(local_int_pos*6)+2],res);

	    res=strtok(NULL,";");
	    if (check_mask(res))
	    {
		local_int_tab[1][(local_int_pos*6)+3]= (char *) calloc(strlen(res)+2,sizeof(char));
		strcpy(local_int_tab[1][(local_int_pos*6)+3],res);
		local_int_tab2[1][(local_int_pos*3)+1] = mask2hex(res);
	    }
	    else
	    {
		syslog(LOG_ERR,"Can't find mask for IP %s in file /etc/shaper/shaper.cfg",res);
//		error("shaperd: brak maski dla numeru IP %s w pliku /etc/shaper/shaper.cfg",res);
		exit(1);
	    }

	    res=strtok(NULL,";");
	    if (check_mask(res))
	    {
		local_int_tab[1][(local_int_pos*6)+4]= (char *) calloc(strlen(res)+2,sizeof(char));
		strcpy(local_int_tab[1][(local_int_pos*6)+4],res);
		local_int_tab2[1][(local_int_pos*3)+2] = mask2hex(res);
	    }
	    else
	    {
		syslog(LOG_ERR,"Can't find mask for IP %s in file /etc/shaper/shaper.cfg",res);
//		error("shaperd: brak maski dla numeru IP %s w pliku /etc/shaper/shaper.cfg.",res);
		exit(1);
	    }

	    res=strtok(NULL,"\n");
	    local_int_tab[1][(local_int_pos*6)+5]= (char *) calloc(strlen(res)+2,sizeof(char *));
	    strcpy(local_int_tab[1][(local_int_pos*6)+5],res);

	    local_int_pos++;
	}
    }
    fclose(fp);
    return(local_int_tab);
}

void free_config(char ***local_int_tab)
{
    int	m,n;

    if (local_int_tab)
    {
	for(m=0;m<local_int_pos;m++)
//	while(local_int_tab[local_int_pos])
	{
	    for(n=1;n<6;n++)
		free(local_int_tab[1][(m*6)+n]);
	}
	free(local_int_tab[1]);
	free(local_int_tab2[1]);
	free(local_int_tab);
	free(local_int_tab2);
	local_int_pos = 0;
    }
}

char ***natstat(char **ignore, char ***config, int firewall_type)
{
    char	source[20],destin[20],sport[10],port[10];
    char	source2[20],destin2[20],sport2[10],port2[10];
    char	s1[20],s2[20],s3[20],s4[20];
    int		timeout;
    int		d1,d2,d3,m,n,fin;
    FILE	*fp;
    char	source_b[20];
    char        source_b2[20];
    char	destin_b[20];
    char	loc_ip_b[20];
    char	gat_ip_b[20];
    unsigned int	source_i;
    unsigned int	destin_i;
    unsigned int	source2_i;
    unsigned int	destin2_i;
    int		source_t;
    int		destin_t;
    unsigned int	loc_ip_i;
    unsigned int	gat_ip_i;
    char	***natstat_tab = NULL;
    char	bufor[1000];
    char	ignorebuf[200];
    int		ignbuf = 0;

//    if ((firewall_type == 1) && (fp = fopen("/root/ipchains.2.4.x.2","r")))
    if ((firewall_type == 1) && (fp = popen("ipchains -L -M -n 2> /dev/null","r")))
    {
	natstat_tab = (char ***) calloc(3,sizeof(char **));
	natstat_tab[1] = (char **) NULL;
	fgets(bufor,1000,fp);
	fgets(bufor,1000,fp);
	while(fgets(bufor,1000,fp))
	{
	    sscanf(bufor,"%s %4d:%s %s %s %s %s -> %s",s1,&d2,s2,source,destin,sport,s3,port);
	    timeout = d2 - (7200-(TIMEOUT/60));
//	    if (timeout < 0) timeout = (TIMEOUT/60)-d2;
//	    else timeout = 7200-d2;
	    if (timeout < 0 && timeout > ((TIMEOUT/60)-7200)) timeout = (TIMEOUT/60)-d2;
	    else timeout = 7200-d2;
	    if (timeout <= TIMEOUT/60)
	    {
		for(m=0;m<local_int_pos;m++)
		{
		    strcpy(ignorebuf,source);
		    strcat(ignorebuf," ");
		    strcat(ignorebuf,destin);
		    strcat(ignorebuf," ");
		    strcat(ignorebuf,port);
		    strcat(ignorebuf,"$");
		    fin = 0;
		    for(n=1;n<ignore_pos;n++)
			if (strstr(ignorebuf,ignore[n])) fin = 1;

		    if (! fin)
		    {
			natstat_tab[1] = (char **) realloc((char **)natstat_tab[1], sizeof(char *)*((natstat_tab_pos+1)*4));
		    	natstat_tab[1][(natstat_tab_pos*4)+1] = (char *) calloc(strlen(source)+2,sizeof(char));
		    	strcpy(natstat_tab[1][(natstat_tab_pos*4)+1],source);
		    	natstat_tab[1][(natstat_tab_pos*4)+2] = (char *) calloc(strlen(destin)+2,sizeof(char));
		    	strcpy(natstat_tab[1][(natstat_tab_pos*4)+2],destin);
		    	natstat_tab[1][(natstat_tab_pos*4)+3] = (char *) calloc(strlen(port)+2,sizeof(char));
		    	strcpy(natstat_tab[1][(natstat_tab_pos*4)+3],port);
    
    			if (debug > 1) printf("natstat: %s %s %s\n",
			    natstat_tab[1][(natstat_tab_pos*4)+1],
    			    natstat_tab[1][(natstat_tab_pos*4)+2],
			    natstat_tab[1][(natstat_tab_pos*4)+3]);
			natstat_tab_pos++;
		    }
		}
	    }
	}
	pclose(fp);
	return(natstat_tab);
    }
    else
    if ((firewall_type == 2) && (fp = fopen(IP_CONNTRACK,"r")))
    {
        natstat_tab = (char ***) calloc(3,sizeof(char **));
        natstat_tab[1] = (char **) NULL;
        while(fgets(bufor,1000,fp))
        {
    	    sscanf(bufor,"%s %d %d %s src=%s dst=%s sport=%s dport=%s src=%s dst=%s sport=%s dport=%s [ASSURED] use=%d",
			s1,&d1,&d2,s2,source,destin,sport,port,source2,destin2,sport2,port2,&d3);
	    timeout=432000-d2;
	    if (strcmp(bufor,"127.0.0.1")
		&& ! strcmp(s2,"ESTABLISHED")
	        && timeout <= TIMEOUT)
	    {
		for(m=0;m<local_int_pos;m++)
		{
		    source_i = ip2hex2(source) & local_int_tab2[1][(m*3)+1];
		    destin_i = ip2hex2(destin) & local_int_tab2[1][(m*3)+2];
		    source2_i = ip2hex2(source2) & local_int_tab2[1][(m*3)+1];
		    destin2_i = ip2hex2(destin2) & local_int_tab2[1][(m*3)+2];
		    loc_ip_i = ip2hex2(config[1][(m*6)+3]) & local_int_tab2[1][(m*3)+1];
		    gat_ip_i = ip2hex2(config[1][(m*6)+4]) & local_int_tab2[1][(m*3)+2];
		    ignbuf = 0;
	    	    if ((source_i == loc_ip_i)
		    && (destin_i != gat_ip_i)
	    	    && (destin_i != loc_ip_i))
		    {
/*
			if (debug > 1) printf("%s %s %s %s\n",
				loc_ip_b,
				source_b,
				gat_ip_b,
				destin_b);
*/
			// emulacja grep'a
			strcpy(ignorebuf,source);
			strcat(ignorebuf," ");
			strcat(ignorebuf,destin);
			strcat(ignorebuf," ");
			strcat(ignorebuf,port);
			strcat(ignorebuf,"$");
			ignbuf = 1;
		    }
		    else
		    if ((source2_i == loc_ip_i)
		    && (destin2_i != gat_ip_i)
		    && (destin2_i != loc_ip_i))
		    {
			strcpy(ignorebuf,source2);
			strcat(ignorebuf," ");
			strcat(ignorebuf,destin2);
			strcat(ignorebuf," ");
			strcat(ignorebuf,port2);
			strcat(ignorebuf,"$");
			ignbuf = 2;
		    }
		    if (ignbuf)
		    {
			fin = 0;
			for(n=1;n<ignore_pos;n++)
		    	    if (strstr(ignorebuf,ignore[n])) fin = 1;

			if (! fin)
			{
		    	    natstat_tab[1] = (char **) realloc((char **)natstat_tab[1], sizeof(char *)*((natstat_tab_pos+1)*4));
			    if (ignbuf == 1)
			    {
		    		natstat_tab[1][(natstat_tab_pos*4)+1] = (char *) calloc(strlen(source)+2,sizeof(char));
    	    			strcpy(natstat_tab[1][(natstat_tab_pos*4)+1],source);
		    		natstat_tab[1][(natstat_tab_pos*4)+2] = (char *) calloc(strlen(destin)+2,sizeof(char));
		    		strcpy(natstat_tab[1][(natstat_tab_pos*4)+2],destin);
		    		natstat_tab[1][(natstat_tab_pos*4)+3] = (char *) calloc(strlen(port)+2,sizeof(char));
		    		strcpy(natstat_tab[1][(natstat_tab_pos*4)+3],port);
			    }
			    else
			    {
		    		natstat_tab[1][(natstat_tab_pos*4)+1] = (char *) calloc(strlen(source2)+2,sizeof(char));
    	    			strcpy(natstat_tab[1][(natstat_tab_pos*4)+1],source2);
		    		natstat_tab[1][(natstat_tab_pos*4)+2] = (char *) calloc(strlen(destin2)+2,sizeof(char));
		    		strcpy(natstat_tab[1][(natstat_tab_pos*4)+2],destin2);
		    		natstat_tab[1][(natstat_tab_pos*4)+3] = (char *) calloc(strlen(port2)+2,sizeof(char));
		    		strcpy(natstat_tab[1][(natstat_tab_pos*4)+3],port2);
			    }
    			    if (debug) printf("natstat: %s %s %s\n",
				natstat_tab[1][(natstat_tab_pos*4)+1],
    			        natstat_tab[1][(natstat_tab_pos*4)+2],
			        natstat_tab[1][(natstat_tab_pos*4)+3]);
			    natstat_tab_pos++;
			}
		    }
		}
	    }
	}
	fclose(fp);
        return(natstat_tab);
    }
    else
    {
	syslog(LOG_ERR,"Can't read from /proc/net/ip_conntrack file");
//	error("Blad: nie moge odczytac /proc/net/ip_conntrack");
	exit(1);
    }
}

void free_natstat_tab(char ***natstat_tab)
{
    int		m,n;
    
    m=0;
    for(m=0;m<natstat_tab_pos;m++)
//    for(m=natstat_tab_pos-1;m>=0;m--)
    {
//	printf("free natstat: ");
//        for(n=1;n<4;n++)
        for(n=3;n>0;n--)
	{
//	    printf("%s ", natstat_tab[1][(m*4)+n]);
    	    free((char *) natstat_tab[1][(m*4)+n]);
	}
//	printf("\n");
    }
    free(natstat_tab[1]);
    free(natstat_tab);
    natstat_tab_pos = 0;
}

char **read_ignore(char *fname)
{
    FILE        *fp;
    char	*res;
    int         m;
    char	buft[100];
    char	**ignore_tab;

    m = numlines(fname);
    ignore_tab = (char **) calloc(m+2,sizeof(char *));
    fp = fopen(fname,"r");
    while ((res=strtok(fgets(buft,100,fp),"\n")))
    {
	ignore_tab[ignore_pos] = (char *) calloc(strlen(buft)+2,sizeof(char));
	strcpy(ignore_tab[ignore_pos],buft);
	if (debug > 1) printf("ignore: %s\n",ignore_tab[ignore_pos]);
	ignore_pos++;
    }
    fclose(fp);
    return(ignore_tab);
}

void free_ignore(char **ignore_tab)
{
    int		n;

    for(n=1;n<ignore_pos;n++)
    {
	free(ignore_tab[n]);
    }
    free(ignore_tab);
    ignore_pos = 1;
}

char **get_masqip(char ***natstat_tab)
{
    char	**masqip = NULL;
    char	*ip;
    int		m,n,fin;

//    masqip = (char **) calloc(iplista_pos+2,sizeof(char *));
    masqip = (char **) malloc(sizeof(char *)*iplista_pos+2);
    for(n=0;n<natstat_tab_pos;n++)
    {
	ip = natstat_tab[1][(n*4)+1];
	fin = 0;
	for(m=1;m<masqip_pos;m++)
	    if (! strcmp(ip,masqip[m]) && ! strcmp(masqip[m],ip)) fin = 1;
	if (! fin)
	{
	    masqip[masqip_pos] = (char *) malloc(sizeof(char)*strlen(ip)+2);
	    strcpy(masqip[masqip_pos++],ip);
	    if (debug) printf("get_masqip: %s\n",ip);
	}
    }
    return(masqip);
}

void free_masqip(char **masqip)
{
    int		n;

    for(n=1;n<masqip_pos;n++)
//    for(n=masqip_pos-1;n>0;n--)
    {
	free(masqip[n]);
    }
    free(masqip);
    masqip_pos = 1;
}

char ***create_speed_table(int ilip)
{
    char	***speed_tab;

    speed_tab = (char ***) calloc(3,sizeof(char **));
    speed_tab[1] = (char **) calloc(ilip*9,sizeof(char *));
    return(speed_tab);
}

void free_speed_table(char ***speed_tab, int pos)
{
    int		m, n;

    if (speed_tab)
    {
	for(n=1;n<8;n++)
	{
	    for(m=0;m<=(pos-1);m++)
//	    while(speed_tab[1][(pos*7)+n])
	    {
		free(speed_tab[1][(m*8)+n]);
//	        pos++;
	    }
	}
	if (speed_tab[1]) free(speed_tab[1]);
	free(speed_tab);
    }
}

void add_ip2speed_table_new(char ***speed_tab, char *ip)
{
    int		pos = 0;

    while(speed_tab[1][(pos*8)+1])
	pos++;

    speed_tab[1][(pos*8)+1] = (char *) calloc(strlen(ip)+2,sizeof(char));
    strcpy(speed_tab[1][(pos*8)+1],ip);
}

void add_ip2speed_table_sub1(char ***speed_tab, char *ip)
{
    int		pos = 0;

    while(speed_tab[1][(pos*8)+2])
	pos++;

    speed_tab[1][(pos*8)+2] = (char *) calloc(strlen(ip)+2,sizeof(char));
    strcpy(speed_tab[1][(pos*8)+2],ip);
}

void add_ip2speed_table_nop1(char ***speed_tab, char *ip)
{
    int		pos = 0;

    while(speed_tab[1][(pos*8)+3])
	pos++;

    speed_tab[1][(pos*8)+3] = (char *) calloc(strlen(ip)+2,sizeof(char));
    strcpy(speed_tab[1][(pos*8)+3],ip);
}

void add_ip2speed_table_add1(char ***speed_tab, char *ip)
{
    int		pos = 0;

    while(speed_tab[1][(pos*8)+4])
	pos++;

    speed_tab[1][(pos*8)+4] = (char *) calloc(strlen(ip)+2,sizeof(char));
    strcpy(speed_tab[1][(pos*8)+4],ip);
}

void add_ip2speed_table_sub2(char ***speed_tab, char *ip)
{
    int		pos = 0;

    while(speed_tab[1][(pos*8)+5])
	pos++;

    speed_tab[1][(pos*8)+5] = (char *) calloc(strlen(ip)+2,sizeof(char));
    strcpy(speed_tab[1][(pos*8)+5],ip);
}

void add_ip2speed_table_nop2(char ***speed_tab, char *ip)
{
    int		pos = 0;

    while(speed_tab[1][(pos*8)+6])
	pos++;

    speed_tab[1][(pos*8)+6] = (char *) calloc(strlen(ip)+2,sizeof(char));
    strcpy(speed_tab[1][(pos*8)+6],ip);
}

void add_ip2speed_table_add2(char ***speed_tab, char *ip)
{
    int		pos = 0;

    while(speed_tab[1][(pos*8)+7])
	pos++;

    speed_tab[1][(pos*8)+7] = (char *) calloc(strlen(ip)+2,sizeof(char));
    strcpy(speed_tab[1][(pos*8)+7],ip);
}

char *get_speed_table_new(char ***speed_tab, int pos)
{
    char	*nil = "\0";
    char	*res;

    if (res = speed_tab[1][((pos-1)*8)+1]) return(res);
    else return(nil);
}

char *get_speed_table_sub1(char ***speed_tab, int pos)
{
    char	*nil = "\0";
    char	*res;

    if (res = speed_tab[1][((pos-1)*8)+2]) return(res);
    else return(nil);
}

char *get_speed_table_nop1(char ***speed_tab, int pos)
{
    char	*nil = "\0";
    char	*res;

    if (res = speed_tab[1][((pos-1)*8)+3]) return(res);
    else return(nil);
}

char *get_speed_table_add1(char ***speed_tab, int pos)
{
    char	*nil = "\0";
    char	*res;

    if (res = speed_tab[1][((pos-1)*8)+4]) return(res);
    else return(nil);
}

char *get_speed_table_sub2(char ***speed_tab, int pos)
{
    char	*nil = "\0";
    char	*res;

    if (res = speed_tab[1][((pos-1)*8)+5]) return(res);
    else return(nil);
}

char *get_speed_table_nop2(char ***speed_tab, int pos)
{
    char	*nil = "\0";
    char	*res;

    if (res = speed_tab[1][((pos-1)*8)+6]) return(res);
    else return(nil);
}

char *get_speed_table_add2(char ***speed_tab, int pos)
{
    char	*nil = "\0";
    char	*res;

    if (res = speed_tab[1][((pos-1)*8)+7]) return(res);
    else return(nil);
}

char *get_ip_iface(char *ip)
{
    int		n;

    for(n=1;n<iplista_pos;n++)
    {
	if (! strcmp(iplista[n],ip) && ! strcmp(ip,iplista[n]))
	{
	    return(iplista_iface[n]);
	}
    }
    return("\0");
}

unsigned long int get_ip_lospeed(char *ip)
{
    int		n;

    for(n=1;n<iplista_pos;n++)
    {
	if (! strcmp(iplista[n],ip) && ! strcmp(ip,iplista[n]))
	{
	    return(iplista_lo[n]);
	}
    }
    return(100);
}

unsigned long int get_ip_hispeed(char *ip)
{
    int		n;

    for(n=1;n<iplista_pos;n++)
    {
	if (! strcmp(iplista[n],ip) && ! strcmp(ip,iplista[n]))
	{
	    return(iplista_hi[n]);
	}
    }
    return(100);
}

int get_config_by_ip(char ***config, char *ip)
{
    int		n;
    char	*iface;

    iface = get_ip_iface(ip);
    for(n=0;n<local_int_pos;n++)
    {
	if (! strcmp(iface,config[1][(n*6)+1])
	    && ! strcmp(config[1][(n*6)+1],iface))
	    return(n);
    }
    return(-1);
}

char *check_mask(char *ip)
{
    char	mask[20];
    int		d1,d2,d3,d4;
    int		n = 0;
    int		no_dots = 0;
    int		m;
    
    if (! strchr(ip,'/')) return(NULL);

    while(ip[n++] != '/')
	;

    if (! strlen(ip+n)) return(NULL);

    m = n;
    while(ip[m])
	if (ip[m++] == '.') no_dots++;

    if (no_dots == 3)
    {
        sscanf(ip+n,"%3d.%3d.%3d.%3d",&d1,&d2,&d3,&d4);    
        if (d1<0 || d1>255 || d2<0 || d2>255 | d3<0 || d3>255 || d4<0 || d4>255)
	    return(NULL);
	else return(ip+n);
    }
    else
    if (no_dots>3 || (no_dots>0 && no_dots<3)) return(NULL);
    else
    if (atoi(ip+n)<=32 || atoi(ip+n)>=0) return(ip+n);

    return(NULL);
}


int ip2hex(char *ip)
{
    char	buf[20];
    int		a,b,c,d;
    int		res;

    sscanf(ip,"%d.%d.%d.%d",&a,&b,&c,&d);
    if (a>=0 && a<=255
	&& b>=0 && b<=255
	&& c>=0 && c<=255
	&& d>=0 && d<=255)
    {
	sprintf(buf,"%02x%02x%02x%02x",a,b,c,d);
	sscanf(buf,"%x",&res);
	return(res);
    }
    else return(0);
}

int ip2hex2(char *ip)
{
    char	buf[20];
    int		a,b,c,d;
    int		res;

    sscanf(ip,"%d.%d.%d.%d",&a,&b,&c,&d);
    if (a>=0 && a<=255
	&& b>=0 && b<=255
	&& c>=0 && c<=255
	&& d>=0 && d<=255)
    {
	res = (a*256+b)*65536+(c*256+d);
	return(res);
    }
    else return(0);
}

char *hex2ip(int ip)
{
    static char	res[20];
    char	ipt[20];
    int		a,b,c,d;

    sprintf(ipt,"%x",ip);
    sscanf(ipt,"%2x%2x%2x%2x",&a,&b,&c,&d);
    sprintf(res,"%d.%d.%d.%d",a,b,c,d);
    return(res);
}

char *hex2ip2(char *ip)
{
    static char	res[20];
    char	ipt[20];
    int		a,b,c,d;

    sprintf(ipt,"%s",ip);
    sscanf(ipt,"%2x%2x%2x%2x",&a,&b,&c,&d);
    sprintf(res,"%d.%d.%d.%d",a,b,c,d);
    return(res);
}

char *get_ip_iface2(char ***local_int_tab, char *ip)
{
    int		intip, mip, maska, n;
    char	*mainip;
    int		res;

    mip = ip2hex(ip);
    for(n=0;n<local_int_pos;n++)
    {
	maska = mask2hex(local_int_tab[1][(n*6)+4]);
        intip = ip2hex(local_int_tab[1][(n*6)+4]);
	res = intip & maska;
//	mainip = hex2ip(intip & mask2hex(local_int_tab[1][(n*6)+4]));
	if ((mip & maska) == res) return(hex2ip(res));
//	printf("%s %x %x %s\n",
//	    ip,
//	    mask2hex(local_int_tab[1][(n*6)+4]),
//	    intip & mask2hex(local_int_tab[1][(n*6)+4]),
//	    hex2ip(intip & mask2hex(local_int_tab[1][(n*6)+4])));
    }
    return(NULL);
}

int mask2hex(char *ip)
{
    char	mask[20];
    int		d1,d2,d3,d4;
    int		n = 0;
    int		no_dots = 0;
    int		m;
    int		res = 0;

    if (! strchr(ip,'/')) return(0);

    while(ip[n++] != '/')
	;

    if (! strlen(ip+n)) return(0);

    m = n;
    while(ip[m])
	if (ip[m++] == '.') no_dots++;

    if (no_dots == 3)
    {
        sscanf(ip+n,"%3d.%3d.%3d.%3d",&d1,&d2,&d3,&d4);    
        if (d1<0 || d1>255 || d2<0 || d2>255 | d3<0 || d3>255 || d4<0 || d4>255)
	    return(0);
	else return(ip2hex(ip+n));
    }
    else
    if (no_dots>3 || (no_dots>0 && no_dots<3)) return(0);
    else
    if (atoi(ip+n)<=32 || atoi(ip+n)>=0)
    {
	for(m=31;m>=(32-atoi(ip+n));m--)
	    res = res + ldexp(1,m);
	return(res);
    }
    return(0);
}

void create_firewall(char ***config, int firewall_type)
{
    int		pos;
    int		n;
    char	bufor[1000];
    FILE	*fp;

    fflush(NULL);
    if (fp = popen("/bin/sh 2>/dev/null","w"))
    {
	if (firewall_type == 1)
	{
	    sprintf(bufor,"ipchains -M -S %d 10 60\n",TIMEOUT);
	    fputs(bufor,fp);
    	    fputs("ipchains -N shaper\n",fp);
	    fputs("ipchains -I output -j shaper\n",fp);
    	    for(n=1;n<iplista_pos;n++)
	    {
		if ((pos = get_config_by_ip(config,iplista[n])) != -1)
		{
		    sprintf(bufor,"ipchains -A shaper -s ! %s -d %s -j RETURN\n",
			config[1][(pos*6)+3],iplista[n]);
		    fputs(bufor,fp);
		}
	    }
	}
	else
	if (firewall_type == 2)
	{
	    fputs("iptables -N shaper\n",fp);
	    sprintf(bufor,"iptables -I FORWARD -i %s -j shaper\n",internet_iface);
	    fputs(bufor,fp);
    	    for(n=1;n<iplista_pos;n++)
	    {
		if ((pos = get_config_by_ip(config,iplista[n])) != -1)
		{
		    sprintf(bufor,"iptables -A shaper -o %s -s ! %s -d %s -j RETURN\n",
			config[1][(pos*6)+1],config[1][(pos*6)+3],iplista[n]);
		    fputs(bufor,fp);
		}
	    }
	}
	pclose(fp);
    }
}

void delete_firewall(char ***config, int firewall_type)
{
    int		pos;
    int		n;
    char	bufor[1000];
    FILE	*fp;

    fflush(NULL);
    if (fp = popen("/bin/sh 2>/dev/null","w"))
    {
	if (firewall_type == 1)
	{
	    fputs("ipchains -F shaper\n",fp);
	    fputs("ipchains -D output -j shaper\n",fp);
	    fputs(bufor,fp);
	    fputs("ipchains -X shaper\n",fp);
	}
	else
	if (firewall_type == 2)
	{
	    fputs("iptables -F shaper\n",fp);
	    sprintf(bufor,"iptables -D FORWARD -i %s -j shaper\n",internet_iface);
	    fputs(bufor,fp);
	    fputs("iptables -X shaper\n",fp);
	}
	pclose(fp);
    }
}

char ***mkbitrate_user_last(char ***bitrate_user_last, char *ip, long int ispeed)
{
    char	speed[20];
    char	time[20];

    
    sprintf(speed,"%d",ispeed);
//    sprintf(time,"%d",time(NULL));
    if ((ckbitrate_user_last(bitrate_user_last,ip)) == -1)
    {
        if (! bitrate_user_last)
	{
	    bitrate_user_last = (char ***) realloc(bitrate_user_last,sizeof(char **)*5);
	    bitrate_user_last[1] = (char **) NULL;
    	    bitrate_user_last[2] = (char **) NULL;
	}
	bitrate_user_last[1] = (char **) realloc((char **) bitrate_user_last[1],sizeof(char *)*((bitrate_user_last_pos+1)*4));
	bitrate_user_last[2] = (char **) realloc((char **) bitrate_user_last[2],sizeof(char *)*((bitrate_user_last_pos+1)*4));
        bitrate_user_last[1][bitrate_user_last_pos+1]= (char *) calloc(strlen(ip)+2,sizeof(char));
	bitrate_user_last[2][bitrate_user_last_pos+1]= (char *) calloc(strlen(speed)+2,sizeof(char));
	strcpy(bitrate_user_last[1][bitrate_user_last_pos+1],ip);
        strcpy(bitrate_user_last[2][bitrate_user_last_pos+1],speed);
	bitrate_user_last_pos++;
    }
    else
    {
	int	n;

	for(n=0;n<bitrate_user_last_pos;n++)
    	    if (! strcmp(bitrate_user_last[1][n+1],ip)
		&& ! strcmp(ip,bitrate_user_last[1][n+1]))
		break;
	bitrate_user_last[1][n+1]= (char *) realloc(bitrate_user_last[1][n+1],sizeof(char *)*((n+1)*(strlen(ip)+2)));
	bitrate_user_last[2][n+1]= (char *) realloc(bitrate_user_last[2][n+1],sizeof(char *)*((n+1)*(strlen(speed)+2)));
	strcpy(bitrate_user_last[1][n+1],ip);
        strcpy(bitrate_user_last[2][n+1],speed);
    }
    return(bitrate_user_last);
}

char ***clear_bitrate_user_last(char ***bitrate_user_last, char **masq)
{
    int		m,n;
    int		*iplista_fin;

    iplista_fin = (int *) calloc(iplista_pos,sizeof(int));
    for(n=1;n<iplista_pos;n++)
	iplista_fin[n] = 0;

    for(n=1;n<iplista_pos;n++)
    {
	for(m=1;m<masqip_pos;m++)
        {
	    if ((!strcmp(iplista[n],masq[m]) && (!strcmp(masq[m],iplista[n]))))
		iplista_fin[n] = 1;
	}
    }

    for(n=1;n<iplista_pos;n++)
	if (! iplista_fin[n])
    	    bitrate_user_last = mkbitrate_user_last(bitrate_user_last,iplista[n],0);

    free(iplista_fin);
    return(bitrate_user_last);
}

long int ckbitrate_user_last(char ***bitrate_user_last, char *ip)
{
    int		n;

    if (bitrate_user_last)
	for(n=0;n<bitrate_user_last_pos;n++)
	    if (! strcmp(bitrate_user_last[1][n+1],ip)
		&& ! strcmp(ip,bitrate_user_last[1][n+1]))
		return(atoi(bitrate_user_last[2][n+1]));
    return(-1);
}

void free_bitrate_user_last(char ***bitrate_user_last)
{
    int		n;

    if (bitrate_user_last)
    {
        for(n=0;n<bitrate_user_last_pos;n++)
        {
    	    free(bitrate_user_last[1][n+1]);
	    free(bitrate_user_last[2][n+1]);
        }
	free(bitrate_user_last[1]);
        free(bitrate_user_last[2]);
	free(bitrate_user_last);
    }
}

void write_bitrate_user_last(char ***bitrate_user_last, char *fname)
{
    FILE	*fw;
    int		n;
    int		fd;
    char	buf[200];
    long int	speed;

    if (fw = fopen(fname,"w"))
    {
//	printf("pos %d\n",bitrate_user_last_pos);
	for(n=0;n<bitrate_user_last_pos;n++)
	{
	    speed = strtol(bitrate_user_last[2][n+1],NULL,10);
	    if (speed > 0)
	    {
		sprintf(buf,"%s %ld %s\n",bitrate_user_last[1][n+1],speed,speed_ext);
		fputs(buf,fw);
	    }
	}
//	close(fd);
	fclose(fw);
    }
}

static int daemonize(void)
{
    int fd;
    int pid;
    char buf[100];

    fprintf(stderr, "daemonizing...");

    if ((pid = fork()) < 0)
    {
	fprintf(stderr, " failed!\n");
	return -1;
    }
    else if (pid != 0)
    {
	unlink(PIDFILE);
	if ((fd = open(PIDFILE, O_CREAT|O_WRONLY, 0644)) >= 0)
	{
	    sprintf(buf,"%d\n",pid);
	    write(fd,buf,strlen(buf));
	    close(fd);
	}
	fprintf(stderr, " done. PID=%d\n", pid);
	exit(0);
    }

    // parent is gone now
    setsid();   // ditch controlling terminal, become session leader
    chdir("/"); // new working directory (don't prevent unmounting a FS)
    umask(0);   // clear inherited file creation mask

    // set up logging facility
    openlog("shaperd", LOG_CONS, LOG_DAEMON);
    syslog(LOG_INFO, "starting: LAN Traffic shaper %s Grzegorz Fitrzyk (sp9wun@poczta.onet.pl)",VERSION);
    syslog(LOG_INFO, "running, PID=%d", getpid());

    return 0;
}

int check_linux(void)
{
    FILE	*fp;
    char	buf[1000];
    char	*p,*q;
    int		a = 0;
    int		b = 0;
    
    if (fp = fopen("/proc/version","r"))
    {
	while(fgets(buf,1000,fp))
	{
	    if (strstr(buf,"Linux version"))
	    {
		p = strtok(buf," ");
		p = strtok(NULL," ");
		p = strtok(NULL," ");
		p = strtok(p,".");
		q = strtok(NULL,".");
		a = atoi(p);
		b = atoi(q);
//		printf("wersja: %d %d\n",a,b);
		if (a == 2)
		{
		    fclose(fp);
		    return(b);
		}
	    }
	}
	fclose(fp);
    }
    return(0);
}

int check_ipchains(void)
{
    FILE	*fp;
    char	buf[1000];

    if (fp = popen("ipchains -L -n 2> /dev/null","r"))
    {
	fgets(buf,1000,fp);
	pclose(fp);
	if (strstr(buf,"Chain input")) return(1);
	else return(0);	
    }
}

int check_iptables(void)
{
    FILE	*fp;
    char	buf[1000];

    if (fp = popen("iptables -L -n 2> /dev/null","r"))
    {
	fgets(buf,1000,fp);
	pclose(fp);
	if (strstr(buf,"Chain INPUT")) return(1);
	else return(0);
    }
}

void create_ipchains(char ***config)
{
    int		pos;
    int		n;
    char	bufor[1000];
    FILE	*fp;

    fflush(NULL);
    if (fp = popen("/bin/sh 2>/dev/null","w"))
    {
	fputs("ipchains -N shaper\n",fp);
	sprintf(bufor,"ipchains -I output -i %s -j shaper\n",internet_iface);
	fputs(bufor,fp);
        for(n=1;n<iplista_pos;n++)
	{
	    if ((pos = get_config_by_ip(config,iplista[n])) != -1)
	    {
		sprintf(bufor,"ipchains -A shaper -i %s -s ! %s -d %s -j RETURN\n",
		    config[1][(pos*6)+1],config[1][(pos*6)+3],iplista[n]);
		fputs(bufor,fp);
	    }
	}
	pclose(fp);
    }
}

void delete_ipchains(char ***config)
{
    int		pos;
    int		n;
    char	bufor[1000];
    FILE	*fp;

    fflush(NULL);
    if (fp = popen("/bin/sh 2>/dev/null","w"))
    {
	fputs("ipchains -F shaper\n",fp);
	sprintf(bufor,"ipchains -D output -i %s -j shaper\n",internet_iface);
	fputs(bufor,fp);
	fputs("ipchains -X shaper\n",fp);
	pclose(fp);
    }
}

int is_shaperd_running(void)
{
    FILE	*fp;
    char	buft[1000];
    int		fin = 0;

// stara metoda kontroli czy shaperd jest juz uruchomiony
/*
    if (fp = popen("ps aux","r"))
    {
	while(fgets(buft,1000,fp))
	{
	    if (strstr(buft,"/shaperd")) fin++;
	}
	pclose(fp);
	return(fin-1);
    }
    return(0);
*/
// nowa metoda kontroli czy shaperd jest juz uruchomiony
    if (fp = fopen(PIDFILE,"r"))
    {
	fclose(fp);
	return(1);
    }
    else return(0);
}

int is_iproute2(void)
{
    FILE	*fp, *fq;
    char	buft[1000];
    int		res = 0;

    fflush;
    if (fp = popen("tc 2>/tmp/tc.tmp\n","r"))
    {
	pclose(fp);
	if (fq = fopen("/tmp/tc.tmp","r"))
	{
	    while (fgets(buft,1000,fq))
	    {
		if (strstr(buft,"sage: tc [")) res = 1;
	    }
	    fclose(fq);
	    fputs("rm -f /tmp/tc.tmp\n",fp);
	}
    }
    return(res);
}

int is_qos(void)
{
    FILE	*fp, *fq;
    char	buft[1000];
    int		res = 0;

    fflush;
    if (fp = popen("tc -s qdisc > /dev/null 2>/tmp/tc.tmp\n","w"))
    {
	pclose(fp);
	if (fq = fopen("/tmp/tc.tmp","r"))
	{
	    fgets(buft,1000,fq);
	    fclose(fq);
	    if (strstr(buft,"RTNETLINK error")) res = 0;
	    else res = 1;
	    fputs("rm -f /tmp/tc.tmp\n",fp);
	}
    }
    return(res);
}

void shownat(char ***natstat_tab)
{
    int		n;

    for(n=0;n<natstat_tab_pos;n++)
    	printf("%s %s %s\n",
	    natstat_tab[1][(n*4)+1],
	    natstat_tab[1][(n*4)+2],
	    natstat_tab[1][(n*4)+3]);
    return;
}

char *lower(char *string)
{
    static char	res[1000];
    int		n = 0;

    while((res[n] = tolower(string[n])) && n<999)
	n++;

    return(res);
}

void exit_handler()
{
    exit_signal = 1;
}
